#include "MeshTree.h"
#include "AMF_File.h"
#include <algorithm>

MeshTreeNode* MeshTreeNode::pStaticThis[AMF_MAX_THREADS]={NULL};

MeshTreeNode& MeshTreeNode::operator=(const MeshTreeNode& m)
{
	Parent = NULL;
	pMesh = NULL;
	pMat=NULL;

	Clear();

	if (m.Parent) Parent = m.Parent;
	
	SubNodes=m.SubNodes;

	for (std::vector<MeshTreeNode>::iterator it = SubNodes.begin(); it != SubNodes.end(); it++)	it->Parent=this;

	if (m.pMesh){
		pMesh = new CMeshSlice;
		*pMesh = *m.pMesh;
	}

	Translated = m.Translated;
	Rotated = m.Rotated;
	GlobalScaled = m.GlobalScaled;
	Min = m.Min;
	Max = m.Max;

	_RotatedInv = m._RotatedInv;

	LocalMin = m.LocalMin;
	LocalMax = m.LocalMax;

	pAmfConst = m.pAmfConst;
	pAmfObject = m.pAmfObject;
	pAmfMesh = m.pAmfMesh;
	pAmfVolume = m.pAmfVolume;

	pMat = m.pMat;
	return *this;
}

void MeshTreeNode::Clear()
{
	SubNodes.clear();
	Translated=Vec3D();
	Rotated=CQuat();
	_RotatedInv = CQuat();
	GlobalScaled=Vec3D(1.0, 1.0, 1.0);
	Min=Vec3D();
	Max=Vec3D();
	LocalMin=Vec3D();
	LocalMax=Vec3D();

	pMat = NULL;

	if (pMesh) {delete pMesh; pMesh=NULL;}

}


MeshTreeNode* MeshTreeNode::AddSubNode(Vec3D AddTranslate, CQuat AddRotate)
{
//	LocalTranslated = AddTranslate;
//	LocalRotated = AddRotate;

	SubNodes.push_back(MeshTreeNode(this));
	MeshTreeNode* pSubNode = &SubNodes.back();

	//Add any additional translation/reotation. order:
	//1) rotate local rotation
	//2) translate local translation
	//3) rotate current rotation (and rotate local translation...
	//4) translate current translation

	CQuat QAddTranslate = CQuat(AddTranslate);
	QAddTranslate = Rotated*QAddTranslate*Rotated.Conjugate(); //rotate previous offsets in proper order

	pSubNode->SetRotated(Rotated*AddRotate); //keep track of total rotation (rotate this amount, then everything previous
	pSubNode->Translated = QAddTranslate.ToVec() + Translated; //add in this offset
	
	return pSubNode;
}

void MeshTreeNode::DrawGL(bool ShowEdges){ //draws the mesh tree...
	if (pMesh){
		pMesh->SetDrawEdges(ShowEdges);
		pMesh->Draw();
	}
	else {
		for (std::vector<MeshTreeNode>::iterator it = SubNodes.begin(); it != SubNodes.end(); it++) it->DrawGL(ShowEdges);
	}

}

Vec3D MeshTreeNode::LocalCoord(Vec3D& GlobalCoord){ //inverts the transformations on this one to transform global coordinate into local one...
//	CQuat ToRotate = Rotated.Inverse();
	return (GlobalCoord-Translated).Rot(_RotatedInv);
} 

void MeshTreeNode::ListMeshNodes (std::vector<MeshTreeNode*>* pNodesWithMeshes) //adds any nodes that contain meshes to this list
{
	if (pMesh) (*pNodesWithMeshes).push_back(this); //this one has a mesh! (if so it should have no subnodes)
	else { //else if it has children, add them to the list...
		for (int i=0; i<(int)SubNodes.size(); i++) SubNodes[i].ListMeshNodes(pNodesWithMeshes); 
	}
}

bool MeshTreeNode::RenderTriMesh(nAmf* pAmfIn, nObject* pObjIn, nMesh* pMeshIn, nVolume* pVolIn, unsigned int CurColorViewIn, int NumSubDiv)
{
	ColorView CurColorView = (ColorView)CurColorViewIn;
	std::vector<inds> AddedTextures; //keep a list of which AMF tex ID's we've added as textures to mesh class the index of these should match the index in the Textures[] array in the mesh

	//create a list of length number of vertices in AMF with number of edge elements touching this vertex
	int NumVerts = (int)pMeshIn->Vertices.VertexList.size();
	int* VertNumEdges = new int[NumVerts];
	for (int i=0; i<NumVerts; i++) VertNumEdges[i]=0; //set to zeros
	for (std::vector<nEdge>::iterator it = pMeshIn->Vertices.EdgeList.begin(); it != pMeshIn->Vertices.EdgeList.end(); it++){
		VertNumEdges[it->v1]++;
		VertNumEdges[it->v2]++;
	}

	//make this a Mesh end node!
	pMesh = new CMeshSlice();
	SubNodes.clear(); //no subnodes allowed if we have a mesh here!

	pMat = pAmfIn->GetMaterialByID(pVolIn->aMaterialID); //set the material pointer for getting interior material info (or sets to null if no material)
	if (pMat) pMesh->pGetInsideColor = &GetInsideColor; //set up interior color callback (this is the one thing CMeshSlice needs to know that it does not store locally)
	

	CVertex v[3];
	nVertex* pTriVerts[3];
	nVertex* pCurVert;
	nColor* pCurColor;
	for (std::vector<nTriangle>::iterator it = pVolIn->Triangles.begin(); it != pVolIn->Triangles.end(); it++){ //Triangles
		pTriVerts[0] = &(pMeshIn->Vertices.VertexList[it->v1]);
		pTriVerts[1] = &(pMeshIn->Vertices.VertexList[it->v2]);
		pTriVerts[2] = &(pMeshIn->Vertices.VertexList[it->v3]);

		//Add vertex colors and locations
		for(int i=0; i<3; i++){
			pCurVert = pTriVerts[i];
			Vec3D ThisLoc = Vec3D(pCurVert->GetX(), pCurVert->GetY(), pCurVert->GetZ()); 

			//Get the color this vertex should have! (always set a color)
			pCurColor = NULL;
			switch (CurColorView){
			case CV_TRI: case CV_TRICOLOR: case CV_TRITEX: if (it->ColorExists) pCurColor=&it->Color; break;
			case CV_VERT: if (pCurVert->ColorExists) pCurColor = &pCurVert->Color; break;
			case CV_VOL: if (pVolIn->ColorExists) pCurColor=&pVolIn->Color; break;
			case CV_OBJ: if (pObjIn->ColorExists)  pCurColor=&pObjIn->Color; break;
			case CV_MAT: if (pMat && pMat->ColorExists) pCurColor=&pMat->Color; break;
			case CV_ALL:
				//order of precedence: material < object < volume < vertex < triangle 
				if (it->ColorExists) pCurColor = &it->Color;
				else if (pCurVert->ColorExists) pCurColor = &pCurVert->Color;
				else if (pVolIn->ColorExists) pCurColor = &pVolIn->Color;
				else if (pObjIn->ColorExists) pCurColor = &pObjIn->Color;
				else if (pMat != NULL && pMat->ColorExists)pCurColor=&pMat->Color;
				break;
			}
			CColor ThisColor = CColor(0.5, 0.5, 0.5);
			if (pCurColor) ThisColor = nColor2CColor(pCurColor, &ThisLoc);

			if (pCurVert->NormalExists){
				Vec3D ThisNormal = Vec3D(pCurVert->Normal.nX, pCurVert->Normal.nY, pCurVert->Normal.nZ);
				v[i] = CVertex(ThisLoc, ThisNormal, ThisColor);
			}
			else v[i] = CVertex(ThisLoc, ThisColor);

		}
	
		//Add in the triangles! (and textures where needed)
		bool ErrorAdding = false;
		if (it->TexMapExists && (CurColorView == CV_TRITEX || CurColorView == CV_TRI || CurColorView == CV_ALL)){ //if there's a texture map and we want to see textures...
			nTexmap* pMap = it->GetpTexMap();
			int MeshTexIndex = GetMeshTexIndex(pAmfIn, pMap, &AddedTextures);
			if (MeshTexIndex == -1) ErrorAdding = true; //couldn't render texture!

			if(!ErrorAdding){
				try{pMesh->AddFacet(v[0], v[1], v[2], TexMap(MeshTexIndex, pMap->uTex1, pMap->uTex2, pMap->uTex3, pMap->vTex1, pMap->vTex2, pMap->vTex3));}
				catch(std::bad_alloc){ErrorAdding = true;}
			}
		}

		else { //otherwise, no texture on triangle
			try{pMesh->AddFacet(v[0], v[1], v[2]);}
			catch(std::bad_alloc){ErrorAdding = true;}
		}
		//for now just drop if ErrorAdding is true. Eventually pass error message!


		//Add edges
		if (VertNumEdges[it->v1] != 0 || VertNumEdges[it->v2] != 0 || VertNumEdges[it->v3] != 0){ //If any edges touch this triangle..
		//Exhaustive check of edges! (there shouldn't be a ton of these compared to numbers of vertices...)
				
			CLine tmpLine;
			int Ei1, Ei2, Ti1, Ti2, Ti3;
			for (std::vector<nEdge>::iterator jt = pMeshIn->Vertices.EdgeList.begin(); jt != pMeshIn->Vertices.EdgeList.end(); jt++){
				Ei1=jt->v1; Ei2=jt->v2; Ti1=it->v1, Ti2=it->v2, Ti3=it->v3;
				CFacet* pCurFacet = pMesh->GetpFacet(pMesh->GetFacetCount()-1); 
				Vec3D Ta = Vec3D(jt->dx1, jt->dy1, jt->dz1);
				Vec3D Tb = Vec3D(jt->dx2, jt->dy2, jt->dz2);
				Vec3D nTa = -Ta;
				Vec3D nTb = -Tb;  

				//from v0 to v1
				if (Ei1==Ti1 && Ei2==Ti2) AddEdge(pCurFacet, 0, Ta, Tb);
				else if (Ei2==Ti1 && Ei1==Ti2) AddEdge(pCurFacet, 0, nTb, nTa); //switch direct to keep consistent
				//from v1 to v2
				else if (Ei1==Ti2 && Ei2==Ti3) AddEdge(pCurFacet, 1, Ta, Tb);
				else if (Ei2==Ti2 && Ei1==Ti3) AddEdge(pCurFacet, 1, nTb, nTa); //switch direct to keep consistent
				//from v2 to v0
				else if (Ei1==Ti3 && Ei2==Ti1) AddEdge(pCurFacet, 2, Ta, Tb);
				else if (Ei2==Ti3 && Ei1==Ti1) AddEdge(pCurFacet, 2, nTb, nTa); //switch direct to keep consistent
			}
		}
	}

	pMesh->CalcFaceNormals(); //if no normal information is provided we need to calc the normal for the flat triangle.

	for (int i=0; i<NumSubDiv; i++) pMesh->SubdivideMe(); //some inefficiency in doing this before transformations, but makes bounding boxes easier.
	pMesh->SetDrawEdges(true);

	LocalMin = pMesh->GetBBMin();
	LocalMax = pMesh->GetBBMax();

	//rotate/translate everything...
	pMesh->Rotate(Rotated);
	pMesh->Translate(Translated);
	
	Min = pMesh->GetBBMin();
	Max = pMesh->GetBBMax();

	delete[] VertNumEdges;

	//Housekeeping for the tree structure
	GenerateMeshIndexStack();
	BoundsPropegateUp(this);

	return true;
}

void MeshTreeNode::GenerateMeshIndexStack(void)
{
	std::vector<int> IndexStack;
	MeshTreeNode* pCurNode = this;

	while (pCurNode->Parent != NULL){
		for (int i=0; i<(int)pCurNode->Parent->SubNodes.size(); i++){
			if (&(pCurNode->Parent->SubNodes[i]) == pCurNode){
				IndexStack.push_back(i);
				break;
			}
		}
		pCurNode = pCurNode->Parent;
	}

	std::reverse(IndexStack.begin(), IndexStack.end()); //put so parent element is first...
	pMesh->GlNameIndexStack = IndexStack;

}


void MeshTreeNode::AddEdge(CFacet* pFacet, int Edge, Vec3D& T1, Vec3D& T2)
{
	int V1, V2;
	switch (Edge){
	case 0: V1=pFacet->vi[0]; V2=pFacet->vi[1]; break;
	case 1: V1=pFacet->vi[1]; V2=pFacet->vi[2]; break;
	case 2: V1=pFacet->vi[2]; V2=pFacet->vi[0]; break;
	}

	CLine tmpLine = CLine(V1, V2, T1, T2);
	pMesh->AddLine(tmpLine);
	pFacet->HasEdge[Edge] = true;
	pFacet->ei[Edge] = pMesh->GetLineCount()-1;
}


int MeshTreeNode::GetMeshTexIndex(nAmf* pAmfIn, nTexmap* pMapIn, std::vector<inds>* pAddedTexs) //returns the CMesh texture index if we've alreadyy added a texture form these AMF indices. Otherwise returns -1
{
	int aRID = pMapIn->RTexID;
	int aGID = pMapIn->GTexID;
	int aBID = pMapIn->BTexID;
	int aAID = pMapIn->ATexID;

	if (aRID == -1 || aGID == -1 || aBID == -1) return false; //Technically this is allowed, but not handled correctly yet.

	//check and return index if we've already added a texture to this mesh with the same indices...
	int NumAdded = pAddedTexs->size();
	for (int i=0; i<NumAdded; i++){
		if ((*pAddedTexs)[i].ix[0]==aRID && (*pAddedTexs)[i].ix[1]==aGID && (*pAddedTexs)[i].ix[2]==aBID && (*pAddedTexs)[i].ix[3]==aAID) return i;
	}

	//If not found, add a new texture from the AMF texture channels
	//do size checks, make sure all textures are the same size...
	int w[4] = {-1, -1, -1, -1};
	int h[4] = {-1, -1, -1, -1};
	nTexture* pTex = pAmfIn->GetTextureByID(aRID);
	if (pTex) pTex->GetSize(&w[0], &h[0]); else return -1;
	pTex = pAmfIn->GetTextureByID(aGID);
	if (pTex) pTex->GetSize(&w[1], &h[1]); else return -1;
	pTex= pAmfIn->GetTextureByID(aBID);
	if (pTex) pTex->GetSize(&w[2], &h[2]); else return -1;
	if (pMapIn->ATexIDExists) pAmfIn->GetTextureByID(aAID)->GetSize(&w[3], &h[3]);
	if (w[0] != w[1] || w[0] != w[2] || (w[3] != -1 && w[0] != w[3])) return -1; //a width did not match up!
	if (h[0] != h[1] || h[0] != h[2] || (h[3] != -1 && h[0] != h[3])) return -1; //a height did not match up!

	//tile this texture if R, G, or B texture in AMF is tiled...
	bool TileThisOne = (pAmfIn->GetTextureByID(pMapIn->RTexID)->aTiled || pAmfIn->GetTextureByID(pMapIn->GTexID)->aTiled || pAmfIn->GetTextureByID(pMapIn->BTexID)->aTiled);

	//All checks out! add a new texture
	CTexture* NewTexture = pMesh->AddTexture(CTexture());
	inds TheseInds = {{aRID, aGID, aBID, aAID}};
	pAddedTexs->push_back(TheseInds);
	int MeshTexIndex = pMesh->GetTextureCount()-1;

	if (pMapIn->ATexIDExists) NewTexture->LoadData(w[0], h[0], pAmfIn->GetTextureByID(aRID)->BinaryData.data(), pAmfIn->GetTextureByID(aGID)->BinaryData.data(), pAmfIn->GetTextureByID(aBID)->BinaryData.data(), pAmfIn->GetTextureByID(aAID)->BinaryData.data(), TileThisOne);
	else NewTexture->LoadData(w[0], h[0], pAmfIn->GetTextureByID(aRID)->BinaryData.data(), pAmfIn->GetTextureByID(aGID)->BinaryData.data(), pAmfIn->GetTextureByID(aBID)->BinaryData.data(), TileThisOne);

	return MeshTexIndex;
}

CColor MeshTreeNode::nColor2CColor(nColor* ColorIn, Vec3D* pLoc)
{
	if (pLoc) return CColor(ColorIn->GetR(pLoc->x, pLoc->y, pLoc->z), ColorIn->GetG(pLoc->x, pLoc->y, pLoc->z), ColorIn->GetB(pLoc->x, pLoc->y, pLoc->z));
	else return CColor(ColorIn->GetR(), ColorIn->GetG(), ColorIn->GetB());
}

bool MeshTreeNode::GetSlice(CSimpleImage* pImgOut, double ZHeight, Vec3D SurfDepth, double XMin, double XMax, double YMin, double YMax, int XPix, int YPix, int ThreadNum, std::string* pMessage)
{
	if (!pMesh) return false; //no mesh for this node!
	pStaticThis[ThreadNum] = this;
	return pMesh->GetSlice(pImgOut, ZHeight, SurfDepth, XMin, XMax, YMin, YMax, XPix, YPix, ThreadNum, pMessage);
}

void MeshTreeNode::GetInsideColor(Vec3D& OrigLoc, unsigned char* rOut, unsigned char* gOut, unsigned char* bOut, unsigned char* aOut, int ThreadNum) //function to get color based on location (pStaticThis MUST be set correctly for this function to work...
{
	if (ThreadNum >= AMF_MAX_THREADS) return;

	Vec3D TmpPt = pStaticThis[ThreadNum]->LocalCoord(OrigLoc);
	if (pStaticThis[ThreadNum]->pMat)
		pStaticThis[ThreadNum]->pMat->GetColorAt(TmpPt.x, TmpPt.y, TmpPt.z, rOut, gOut, bOut, aOut);
}

void MeshTreeNode::GlobalScaleNode(Vec3D ScaleFactors) //scales the actual meshes (post RenderTriMesh() ) (for global scaling prior to slicing without messing with all the transformations)
{
	if (pMesh){ //if this is a mesh node
		pMesh->Scale(ScaleFactors); //mesh is in global coordinates
		GlobalScaled = GlobalScaled.Scale(ScaleFactors);

		//hacky... previous minimum values will live on, unfortunately.
		Min = pMesh->GetBBMin();
		Max = pMesh->GetBBMax();
		BoundsPropegateUp(this);
	}
	else {
		for (int i=0; i<(int)SubNodes.size(); i++) SubNodes[i].GlobalScaleNode(ScaleFactors); //pass down to all children
	}
}


void MeshTreeNode::BoundsPropegateUp(MeshTreeNode* pSourceMeshNode) //called by a node whenever we add a mesh to that node... propegates the min, maxs, etc. up through the tree...
{
	if (Parent == NULL) return; //if this is the root node, we can stop!

	//absolute max/min
	if (Parent->Min == Vec3D() && Parent->Max == Vec3D()){ //if there's nothing yet, we don't want to assume 0 is a minimum or maximum
		Parent->Min = Min;
		Parent->Max = Max;
	}
	else { //if there's already some min/max info
		if (Min.x < Parent->Min.x) Parent->Min.x = Min.x;
		if (Min.y < Parent->Min.y) Parent->Min.y = Min.y;
		if (Min.z < Parent->Min.z) Parent->Min.z = Min.z;

		if (Max.x > Parent->Max.x) Parent->Max.x = Max.x;
		if (Max.y > Parent->Max.y) Parent->Max.y = Max.y;
		if (Max.z > Parent->Max.z) Parent->Max.z = Max.z;

	}

	//Local max/min
	bool AnyLocalMaxMinInfo = true;
	if (Parent->LocalMin==Vec3D() && Parent->LocalMax==Vec3D()) AnyLocalMaxMinInfo=false;

	if (Parent->Translated == pSourceMeshNode->Translated && Parent->Rotated == pSourceMeshNode->Rotated){ //if no coordinate transformation, LocalMin and LocalMax can be used directly
		if (!AnyLocalMaxMinInfo){ //if there's nothing yet, we don't want to assume 0 is a minimum or maximum
			Parent->LocalMin = LocalMin;
			Parent->LocalMax = LocalMax;
		}
		else { //if there's already some min/max info
			if (LocalMin.x < Parent->LocalMin.x) Parent->LocalMin.x = LocalMin.x;
			if (LocalMin.y < Parent->LocalMin.y) Parent->LocalMin.y = LocalMin.y;
			if (LocalMin.z < Parent->LocalMin.z) Parent->LocalMin.z = LocalMin.z;

			if (LocalMax.x > Parent->LocalMax.x) Parent->LocalMax.x = LocalMax.x;
			if (LocalMax.y > Parent->LocalMax.y) Parent->LocalMax.y = LocalMax.y;
			if (LocalMax.z > Parent->LocalMax.z) Parent->LocalMax.z = LocalMax.z;

		}
	}
	else { //if there was transformation
		////get the relative transformation to go from end nodes coordinate system to this coordinate system (rotate then translate)
		//Vec3D TmpTanslate = Parent->Translated - pSourceMeshNode->Translated;
		//CQuat TmpRotate = pSourceMeshNode->Rotated.Conjugate()*Parent->Rotated;
		////http://www.gamedev.net/topic/423462-rotation-difference-between-two-quaternions/

		//rotations from absolute coordinates of the mesh to the coordinate system of the parent...
//		Vec3D TmpTanslate = Parent->Translated - pSourceMeshNode->Translated;
//		CQuat TmpRotate = pSourceMeshNode->Rotated.Conjugate()*Parent->Rotated;
	
		Vec3D TmpTanslate = -Parent->Translated; // - pSourceMeshNode->Translated;
		CQuat TmpRotate = /*pSourceMeshNode->Rotated.Conjugate()**/Parent->Rotated.Conjugate();

		//could be faster with direct iterator...
		int NumVerts = pSourceMeshNode->pMesh->GetVertexCount();
		CVertex* pCurVert;
		CQuat ThisPos;
		Vec3D FinalPos;
		for (int i=0; i<NumVerts; i++){
			pCurVert = pSourceMeshNode->pMesh->GetpVertex(i);

			//translate, then rotate goint backwards...
			ThisPos = CQuat(pCurVert->v+TmpTanslate);
			ThisPos = TmpRotate*ThisPos*TmpRotate.Conjugate(); //rotate the point...
			FinalPos = ThisPos.ToVec(); //+TmpTanslate;

			if (i==0 && !AnyLocalMaxMinInfo){ //if there's no min max info yet, don't assume zero is a min or max...
				Parent->LocalMin = FinalPos;
				Parent->LocalMax = FinalPos;
			}
			else {
				if (FinalPos.x < Parent->LocalMin.x) Parent->LocalMin.x = FinalPos.x;
				if (FinalPos.y < Parent->LocalMin.y) Parent->LocalMin.y = FinalPos.y;
				if (FinalPos.z < Parent->LocalMin.z) Parent->LocalMin.z = FinalPos.z;

				if (FinalPos.x > Parent->LocalMax.x) Parent->LocalMax.x = FinalPos.x;
				if (FinalPos.y > Parent->LocalMax.y) Parent->LocalMax.y = FinalPos.y;
				if (FinalPos.z > Parent->LocalMax.z) Parent->LocalMax.z = FinalPos.z;
			}

		}
	}

	Parent->BoundsPropegateUp(pSourceMeshNode);


}
